==> jupedsim-0.10.0.tar.gz: jupedsim-0.10.0/libsimulator/src/Stage.hpp <==
/// Copyright © 2012-2022 Forschungszentrum Jülich GmbH
/// SPDX-License-Identifier: LGPL-3.0-or-later
#pragma once

#include "GenericAgent.hpp"
#include "Logger.hpp"
#include "NeighborhoodSearch.hpp"
#include "Point.hpp"
#include "Polygon.hpp"
#include "UniqueID.hpp"

#include <algorithm>
#include <iterator>
#include <limits>
#include <queue>
#include <vector>

class Simulation;

class Stage
{
public:
    virtual ~Stage() = default;
    virtual bool IsCompleted(const GenericAgent& agent) = 0;
    virtual Point Target(const GenericAgent& agent) = 0;
};

class Waypoint : public Stage
{
    Point position;
    double distance;

public:
    Waypoint(Point position_, double distance_);
    ~Waypoint() override = default;
    bool IsCompleted(const GenericAgent& agent) override;
    Point Target(const GenericAgent& agent) override;
};

/// Notifies simulation of all agents that need to be removed at the beginning of the next iteration
class Exit : public Stage
{
    Polygon area;
    std::vector<GenericAgent::ID>& toRemove;

public:
    Exit(Polygon area, std::vector<GenericAgent::ID>& toRemove_);
    ~Exit() override = default;
    bool IsCompleted(const GenericAgent& agent) override;
    Point Target(const GenericAgent& agent) override;
};

class NotifiableWaitingSet : public Stage
{
public:
    enum class WaitingState {
        Active,
        Inactive,
    };

private:
    std::vector<Point> slots;
    jps::UniqueID<Journey> journeyId;
    std::vector<GenericAgent::ID> occupants{};
    WaitingState state{WaitingState::Active};

public:
    NotifiableWaitingSet(std::vector<Point> slots_, jps::UniqueID<Journey> journeyId);
    ~NotifiableWaitingSet() override = default;
    bool IsCompleted(const GenericAgent& agent) override;
    Point Target(const GenericAgent& agent) override;
    void State(WaitingState s);
    WaitingState State() const;
    template <typename T>
    void Update(const NeighborhoodSearch<T>& neighborhoodSearch);
};

template <typename T>
void NotifiableWaitingSet::Update(const NeighborhoodSearch<T>& neighborhoodSearch)
{
    if(state == WaitingState::Inactive) {
        return;
    }
    const auto count_occupants = occupants.size();
    if(count_occupants == slots.size()) {
        return;
    }

    for(size_t index = count_occupants; index < slots.size(); ++index) {
        const auto candidates = neighborhoodSearch.GetNeighboringAgents(slots[index], 2);
        GenericAgent::ID occupant = GenericAgent::ID::Invalid;
        double min_distance = std::numeric_limits<double>::max();
        for(const auto& agent : candidates) {
            if(agent.journeyId == journeyId) {
                if(std::find(std::begin(occupants), std::end(occupants), agent.id) ==
                   std::end(occupants)) {
                    const auto distance = (agent.pos - slots[index]).Norm();
                    if(distance < min_distance) {
                        min_distance = distance;
                        occupant = agent.id;
                    }
                }
            }
        }
        if(occupant != GenericAgent::ID::Invalid) {
            occupants.push_back(occupant);
        } else {
            return;
        }
    }
}

class NotifiableQueue : public Stage
{

private:
    std::vector<Point> slots;
    jps::UniqueID<Journey> journeyId;
    std::vector<GenericAgent::ID> occupants{};
    std::vector<GenericAgent::ID> exitingThisUpdate{};
    size_t popCountOnNextUpdate{};

public:
    NotifiableQueue(std::vector<Point> slots_, jps::UniqueID<Journey> journeyId_);
    ~NotifiableQueue() override = default;
    bool IsCompleted(const GenericAgent& agent) override;
    Point Target(const GenericAgent& agent) override;
    template <typename T>
    void Update(const NeighborhoodSearch<T>& neighborhoodSearch);
    void Pop(size_t count);
};

template <typename T>
void NotifiableQueue::Update(const NeighborhoodSearch<T>& neighborhoodSearch)
{
    exitingThisUpdate.clear();
    if(popCountOnNextUpdate > 0) {
        const auto beginCopy = std::begin(occupants);
        const auto endCopy = beginCopy + popCountOnNextUpdate;
        std::copy(beginCopy, endCopy, std::back_inserter(exitingThisUpdate));
        occupants.erase(beginCopy, endCopy);
        popCountOnNextUpdate = 0;
    }

    const auto count_occupants = occupants.size();
    if(count_occupants == slots.size()) {
        return;
    }

    for(size_t index = count_occupants; index < slots.size(); ++index) {
        const auto candidates = neighborhoodSearch.GetNeighboringAgents(slots[index], 2);
        GenericAgent::ID occupant = GenericAgent::ID::Invalid;
        double min_distance = std::numeric_limits<double>::max();
        for(const auto& agent : candidates) {
            if(agent.journeyId == journeyId) {
                if(std::find(std::begin(occupants), std::end(occupants), agent.id) ==
                   std::end(occupants)) {
                    const auto distance = (agent.pos - slots[index]).Norm();
                    if(distance < min_distance) {
                        min_distance = distance;
                        occupant = agent.id;
                    }
                }
            }
        }
        if(occupant != GenericAgent::ID::Invalid) {
            occupants.push_back(occupant);
            LOG_DEBUG("occupant {} enqueued @ {}", occupant, occupants.size());
        } else {
            return;
        }
    }
}

==> jupedsim-0.11.1.tar.gz: jupedsim-0.11.1/libsimulator/src/Stage.hpp <==
// Copyright © 2012-2023 Forschungszentrum Jülich GmbH
// SPDX-License-Identifier: LGPL-3.0-or-later
#pragma once

#include "GenericAgent.hpp"
#include "Logger.hpp"
#include "NeighborhoodSearch.hpp"
#include "Point.hpp"
#include "Polygon.hpp"
#include "UniqueID.hpp"

#include <algorithm>
#include <iterator>
#include <limits>
#include <queue>
#include <vector>

class Simulation;

class BaseStage;

enum class WaitingSetState {
    Active,
    Inactive,
};

class BaseProxy
{
protected:
    const Simulation* simulation;
    BaseStage* stage;

    BaseProxy(const Simulation* simulation_, BaseStage* stage_)
        : simulation(simulation_), stage(stage_)
    {
    }
    virtual ~BaseProxy() = default;

public:
    size_t CountTargeting() const;
};

class WaypointProxy : public BaseProxy
{
public:
    WaypointProxy(const Simulation* simulation_, BaseStage* stage_) : BaseProxy(simulation_, stage_)
    {
    }
};

class NotifiableWaitingSetProxy : public BaseProxy
{
public:
    NotifiableWaitingSetProxy(const Simulation* simulation_, BaseStage* stage_)
        : BaseProxy(simulation_, stage_)
    {
    }
    void State(WaitingSetState newState);
    WaitingSetState State() const;
    size_t CountWaiting() const;
    const std::vector<GenericAgent::ID>& Waiting() const;
};

class NotifiableQueueProxy : public BaseProxy
{
public:
    NotifiableQueueProxy(const Simulation* simulation_, BaseStage* stage_)
        : BaseProxy(simulation_, stage_)
    {
    }

    size_t CountEnqueued() const;
    const std::vector<GenericAgent::ID>& Enqueued() const;
    void Pop(size_t count);
};

class ExitProxy : public BaseProxy
{
public:
    ExitProxy(const Simulation* simulation_, BaseStage* stage_) : BaseProxy(simulation_, stage_) {}
};

using StageProxy =
    std::variant<WaypointProxy, NotifiableWaitingSetProxy, NotifiableQueueProxy, ExitProxy>;

class BaseStage
{
public:
    using ID = jps::UniqueID<BaseStage>;

protected:
    ID id;

public:
    virtual ~BaseStage() = default;
    virtual bool IsCompleted(const GenericAgent& agent) = 0;
    virtual Point Target(const GenericAgent& agent) = 0;
    virtual StageProxy Proxy(const Simulation* simulation_) = 0;
    ID Id() const { return id; }
};

class Waypoint : public BaseStage
{
    Point position;
    double distance;

public:
    Waypoint(Point position_, double distance_);
    ~Waypoint() override = default;
    bool IsCompleted(const GenericAgent& agent) override;
    Point Target(const GenericAgent& agent) override;
    StageProxy Proxy(const Simulation* simulation_) override;
};

/// Notifies simulation of all agents that need to be removed at the beginning of the next iteration
class Exit : public BaseStage
{
    Polygon area;
    std::vector<GenericAgent::ID>& toRemove;

public:
    Exit(Polygon area, std::vector<GenericAgent::ID>& toRemove_);
    ~Exit() override = default;
    bool IsCompleted(const GenericAgent& agent) override;
    Point Target(const GenericAgent& agent) override;
    StageProxy Proxy(const Simulation* simulation_) override;
};

class NotifiableWaitingSet : public BaseStage
{
    std::vector<Point> slots;
    std::vector<GenericAgent::ID> occupants{};
    WaitingSetState state{WaitingSetState::Active};

public:
    NotifiableWaitingSet(std::vector<Point> slots_);
    ~NotifiableWaitingSet() override = default;
    bool IsCompleted(const GenericAgent& agent) override;
    Point Target(const GenericAgent& agent) override;
    StageProxy Proxy(const Simulation* simulation_) override;
    void State(WaitingSetState s);
    WaitingSetState State() const;
    template <typename T>
    void Update(const NeighborhoodSearch<T>& neighborhoodSearch);
    const std::vector<GenericAgent::ID>& Occupants() const;
};

template <typename T>
void NotifiableWaitingSet::Update(const NeighborhoodSearch<T>& neighborhoodSearch)
{
    if(state == WaitingSetState::Inactive) {
        return;
    }
    const auto count_occupants = occupants.size();
    if(count_occupants == slots.size()) {
        return;
    }

    for(size_t index = count_occupants; index < slots.size(); ++index) {
        const auto candidates = neighborhoodSearch.GetNeighboringAgents(slots[index], 2);
        GenericAgent::ID occupant = GenericAgent::ID::Invalid;
        double min_distance = std::numeric_limits<double>::max();
        for(const auto& agent : candidates) {
            if(agent.stageId == id) {
                if(std::find(std::begin(occupants), std::end(occupants), agent.id) ==
                   std::end(occupants)) {
                    const auto distance = (agent.pos - slots[index]).Norm();
                    if(distance < min_distance) {
                        min_distance = distance;
                        occupant = agent.id;
                    }
                }
            }
        }
        if(occupant != GenericAgent::ID::Invalid) {
            occupants.push_back(occupant);
        } else {
            return;
        }
    }
}

class NotifiableQueue : public BaseStage
{

private:
    std::vector<Point> slots;
    std::vector<GenericAgent::ID> occupants{};
    std::vector<GenericAgent::ID> exitingThisUpdate{};
    size_t popCountOnNextUpdate{};

public:
    NotifiableQueue(std::vector<Point> slots_);
    ~NotifiableQueue() override = default;
    bool IsCompleted(const GenericAgent& agent) override;
    Point Target(const GenericAgent& agent) override;
    StageProxy Proxy(const Simulation* simulation_) override;
    template <typename T>
    void Update(const NeighborhoodSearch<T>& neighborhoodSearch);
    void Pop(size_t count);
    const std::vector<GenericAgent::ID>& Occupants() const;
};

template <typename T>
void NotifiableQueue::Update(const NeighborhoodSearch<T>& neighborhoodSearch)
{
    exitingThisUpdate.clear();
    if(popCountOnNextUpdate > 0) {
        const auto beginCopy = std::begin(occupants);
        const auto endCopy = beginCopy + popCountOnNextUpdate;
        std::copy(beginCopy, endCopy, std::back_inserter(exitingThisUpdate));
        occupants.erase(beginCopy, endCopy);
        popCountOnNextUpdate = 0;
    }

    const auto count_occupants = occupants.size();
    if(count_occupants == slots.size()) {
        return;
    }

    for(size_t index = count_occupants; index < slots.size(); ++index) {
        const auto candidates = neighborhoodSearch.GetNeighboringAgents(slots[index], 2);
        GenericAgent::ID occupant = GenericAgent::ID::Invalid;
        double min_distance = std::numeric_limits<double>::max();
        for(const auto& agent : candidates) {
            if(agent.stageId == id) {
                if(std::find(std::begin(occupants), std::end(occupants), agent.id) ==
                   std::end(occupants)) {
                    const auto distance = (agent.pos - slots[index]).Norm();
                    if(distance < min_distance) {
                        min_distance = distance;
                        occupant = agent.id;
                    }
                }
            }
        }
        if(occupant != GenericAgent::ID::Invalid) {
            occupants.push_back(occupant);
        } else {
            return;
        }
    }
}
